;; Shriram, then Moy, then Robby, then Shriram

; Aries adds begin, begin0 and let's to the transformed source.

; Dependencies:
;   ariesus.ss
;   pretty.ss  [for debugging]

    (unit/sig plt:aries^
      (import [z : zodiac:system^]
	[z:interface : zodiac:interface^])

      (define error-box
	(box #f))

      (define make-improper
	(lambda (combine)
	  (rec improper ;; `rec' is for the name in error messages
	       (lambda (f list)
		 (let improper-loop ([list list])
		   (cond
		     ((null? list) list)
		     ((pair? list) (combine (f (car list))
					    (improper-loop (cdr list))))
		     (else (f list))))))))
      (define improper-map (make-improper cons))
      (define improper-foreach (make-improper (lambda (x y) y)))

      ; Robby's old definition, commented out:
    
      (quote
	(define unparse-read
	  (lambda (read)
	    (cond
	      [(z:improper-list? read)
		(let loop ([l (z:read-object read)])
		  (cond
		    [(null? (cdr l)) (unparse-read (car l))]
		    [else (cons (unparse-read (car l)) (loop (cdr l)))]))]
	      [(z:vector? read)
		(apply vector (map unparse-read (z:read-object read)))]
	      [(z:list? read) (map unparse-read (z:read-object read))]
	      [else (z:read-object read)]))))

      ; Objects that are passed to eval get quoted by M3.  These objects
      ; do not belong in the `read' structure framework.  Hence, if they
      ; are passed to z:sexp->raw, they will error.  Thus, we first check
      ; before sending things there.

      (define read->raw
	(lambda (read)
	  (if (z:zodiac? read)
	    (z:sexp->raw read)
	    read)))

      (define unparse-read read->raw)

      ; We have two kinds of wrapping.  The first, `wrap', checks for
      ; errors in arbitrary code.  In non-tail positions, we have to
      ; also be sure that multiple values were not returned where only
      ; one value was expected.  For this, we use `mv-wrap'.

      (define wrap
	(lambda (zodiac x)
	  (let ([start (z:zodiac-start zodiac)]
		 [finish (z:zodiac-finish zodiac)])
	    `(#%begin (,set-box! ,error-box
			,(z:make-zodiac #f start finish))
	       ,x))))

      (define mv-wrap
	(case-lambda
	  ((body) (mv-wrap body (annotate body)))
	  ((zodiac body)
	    (let ((start (z:zodiac-start zodiac))
		   (finish (z:zodiac-finish zodiac)))
	      `(#%begin0 ,body
		 (,set-box! ,error-box
		   ,(z:make-zodiac #f start finish)))))))

      (define check-for-keyword/both
	(lambda (disallow-procedures?)
	  (lambda (id)
	    (let ([real-id
		   (cond
		     [(z:binding? id) (z:binding-orig-name id)]
		     [(z:top-level-varref? id) (z:varref-var id)]
		     [(z:bound-varref? id)
		      (z:binding-orig-name (z:bound-varref-binding id))]
		     [(z:symbol? id)
		      (z:read-object id)]
		     [else
		      (z:interface:internal-error id
						  "Given in check-for-keyword")])])
	      (when (and (keyword-name? real-id)
			 (or disallow-procedures?
			     (let ([gdv (global-defined-value real-id)])
			       (or (syntax? gdv)
				   (macro? gdv)))))
		(z:interface:static-error id "Invalid use of keyword ~s"
					  real-id))))))
      
      (define check-for-keyword (check-for-keyword/both #t))
      (define check-for-keyword/proc (check-for-keyword/both #f))

      (define paroptarglist->ilist
	(lambda (paroptarglist)
	  (let ((process-args
		  (lambda (element)
		    (if (pair? element)
		      (and (check-for-keyword (car element))
			(list (z:binding-var (car element))
			  (annotate (cdr element))))
		      (and (check-for-keyword element)
			(z:binding-var element))))))
	    (cond
	      ((z:sym-paroptarglist? paroptarglist)
		(process-args (car (z:paroptarglist-vars paroptarglist))))
	      ((z:list-paroptarglist? paroptarglist)
		(map process-args (z:paroptarglist-vars paroptarglist)))
	      ((z:ilist-paroptarglist? paroptarglist)
		(let loop ((vars (map process-args 
				   (z:paroptarglist-vars paroptarglist))))
		  (if (null? (cddr vars))
		    (cons (car vars) (cadr vars))
		    (cons (car vars) (loop (cdr vars))))))
	      (else
		(z:interface:internal-error paroptarglist
		  "Given to paroptarglist->ilist"))))))

      (define arglist->ilist
	(lambda (arglist)
	  (cond
	    ((z:list-arglist? arglist)
	      (z:arglist-vars arglist))
	    ((z:ilist-arglist? arglist)
	      (let loop ((vars (z:arglist-vars arglist)))
		(if (null? (cddr vars))
		  (cons (car vars) (cadr vars))
		  (cons (car vars) (loop (cdr vars))))))
	    ((z:sym-arglist? arglist)
	      (car (z:arglist-vars arglist)))
	    (else
	      (z:interface:internal-error arglist
		"Given to arglist->ilist")))))

      (define the-undefined-value (letrec ((x x)) x))

      (define-struct (undefined struct:exn) (id))
      (define signal-undefined (make-parameter #t))
      (define undefined-error-format
	 "Variable ~s referenced before definition or initialization")

      (define-struct (not-boolean struct:exn) (val))
      (define signal-not-boolean (make-parameter #f))
      (define not-boolean-error-format "if: value ~e is neither #t nor #f")

      (define annotate
	(lambda (expr)
	  (cond
	    [(z:bound-varref? expr)
	     (let ([v (z:varref-var expr)]
		   [real-v (z:binding-orig-name
			    (z:bound-varref-binding expr))])
	       (if (signal-undefined)
		   (wrap expr
			 `(#%if (#%eq? ,v ,the-undefined-value)
				(#%raise (,make-undefined
					  ,(format undefined-error-format real-v)
					  ((#%debug-info-handler))
					  (#%quote ,v)))
				,v))
		   v))]

	    [(z:top-level-varref? expr)
	     (check-for-keyword/proc expr)
	     (let ((v (z:varref-var expr)))
	       (if (signal-undefined)
		   (wrap expr
			 `(#%if (#%eq? ,v ,the-undefined-value)
			   (#%raise (,make-undefined
				     ,(format undefined-error-format v)
				     ((#%debug-info-handler))
				     (#%quote ,v)))
			   ,v))
		   (wrap expr (z:varref-var expr))))]

	    [(z:app? expr)
	      (let* ([aries:app-arg (gensym 'aries:app-arg)]
		      [aries:app-break (gensym 'aries:app-break)]
		      [last-arg (gensym 'last-arg)]
		      [fun-sym (gensym "fun")]
		      [args (map (lambda (x) `(,(gensym "arg")
						,(annotate x)))
			      (z:app-args expr))])
		`(#%let ([,fun-sym ,(annotate (z:app-fun expr))]
			  ,@args)
		   ,(wrap expr `(,fun-sym ,@(map car args)))))]

	    [(z:struct-form? expr)
	      `(#%struct
		 ,(if (z:struct-form-super expr)
		    (list (read->raw (z:struct-form-type expr))
		      (wrap (z:struct-form-super expr)
			(annotate (z:struct-form-super expr))))
		    (read->raw (z:struct-form-type expr)))
		 ,(map read->raw (z:struct-form-fields expr)))]

	    [(z:if-form? expr)
	     (if (signal-not-boolean)
		 (let ([if-test-v (gensym "if-test-v")])
		   `(#%let ((,if-test-v ,(wrap (z:if-form-test expr)
					       (annotate (z:if-form-test expr)))))
			   (#%if (#%boolean? ,if-test-v)
				 (#%if ,if-test-v
				       ,(annotate (z:if-form-then expr))
				       ,(annotate (z:if-form-else expr)))
				 (#%raise (,make-not-boolean
					   (#%format ,not-boolean-error-format
						     ,if-test-v)
					   ((#%debug-info-handler))
					   ,if-test-v)))))
		 `(#%if ,(annotate (z:if-form-test expr))
			,(annotate (z:if-form-then expr))
			,(annotate (z:if-form-else expr))))]

	    [(z:quote-form? expr)
	      `(#%quote ,(unparse-read (z:quote-form-expr expr)))]

	    [(z:begin-form? expr)
	      `(#%begin
		 ,@(map annotate (z:begin-form-bodies expr)))]

	    [(z:begin0-form? expr)
	      `(#%begin0
		 ,@(map annotate (z:begin0-form-bodies expr)))]

	    [(z:let-values-form? expr)
	     (let ([bindings
		    (map (lambda (vars val)
			   (for-each check-for-keyword vars)
			   (let ([name (and (= 1 (length vars))
					    (z:binding-orig-name (car vars)))])
			     `(,(map z:binding-var vars)
			       ,(if name
				    `(#%let ([,name ,(mv-wrap val)])
				      ,name)
				    (mv-wrap val)))))
			 (z:let-values-form-vars expr)
			 (z:let-values-form-vals expr))])
	       `(#%let-values ,bindings
		 ,(annotate (z:let-values-form-body expr))))]

	    [(z:letrec*-values-form? expr)
	      (let ((bindings
		      (map (lambda (vars val)
			     (for-each check-for-keyword vars)
			     (let ([name (and (= 1 (length vars))
					      (z:binding-orig-name (car vars)))])
			       `(,(map z:binding-var vars)
				 ,(if name
				      `(#%let ([,name ,(mv-wrap val)])
					,name)
				      (mv-wrap val)))))
			(z:letrec*-values-form-vars expr)
			(z:letrec*-values-form-vals expr))))
		`(#%letrec*-values ,bindings
		   ,(annotate (z:letrec*-values-form-body expr))))]

	    [(z:define-values-form? expr)
	      `(#%define-values
		 ,(map (lambda (v)
			 (check-for-keyword v)
			 (z:varref-var v))
		    (z:define-values-form-vars expr))
		 ,(mv-wrap (z:define-values-form-val expr)))]

	    [(z:set!-form? expr)
	      (check-for-keyword (z:set!-form-var expr))
	      (let ([g (gensym "set!")])
		`(#%let ([,g ,(mv-wrap (z:set!-form-val expr))])
		   ,(wrap expr 
		      `(#%set! ,(z:varref-var (z:set!-form-var expr))
			 ,g))))]

	    [(z:case-lambda-form? expr)
	      `(#%case-lambda
		 ,@(map (lambda (args body)
			  (let ((args (arglist->ilist args)))
			    (improper-foreach check-for-keyword args)
			    `(,(improper-map z:binding-var args)
			       ,(annotate body))))
		     (z:case-lambda-form-args expr)
		     (z:case-lambda-form-bodies expr)))]

	    [(z:unit-form? expr)
	      (let ((imports (z:unit-form-imports expr))
		     (exports (map (lambda (export)
				     (list (z:varref-var (car export))
				       (z:read-object (cdr export))))
				(z:unit-form-exports expr)))
		     (clauses (map annotate (z:unit-form-clauses expr))))
		(for-each check-for-keyword imports)
		`(#%unit
		   (import ,@(map z:binding-var imports))
		   (export ,@exports)
		   ,@clauses))]

	    [(z:compound-unit-form? expr)
	      (let ((imports (map z:binding-var
			       (z:compound-unit-form-imports expr)))
		     (links (z:compound-unit-form-links expr))
		     (exports (z:compound-unit-form-exports expr)))
		(let
		  ((links
		     (map
		       (lambda (link-clause)
			 (let ((tag (read->raw (car link-clause)))
				(sub-unit (mv-wrap (cadr link-clause)))
				(imports
				  (map (lambda (import)
					 (if (z:lexical-varref? import)
					   (z:varref-var import)
					   `(,(read->raw (car import))
					      ,(read->raw (cdr import)))))
				    (cddr link-clause))))
			   `(,tag (,sub-unit ,@imports))))
		       links))
		    (exports
		      (map
			(lambda (export-clause)
			  `(,(read->raw (car export-clause))
			     (,(read->raw (cadr export-clause))
			       ,(read->raw (cddr export-clause)))))
			exports)))
		  (let ((e `(#%compound-unit
			      (import ,@imports)
			      (link ,@links)
			      (export ,@exports))))
		    e)))]

	    [(z:invoke-unit-form? expr)
	      `(#%invoke-unit ,(mv-wrap (z:invoke-unit-form-unit expr))
		 ,@(map z:varref-var
		     (z:invoke-unit-form-variables expr)))]

	    [(z:invoke-open-unit-form? expr)
	      (let ((name-spec (z:invoke-open-unit-form-name-specifier
				 expr))
		     (unit (mv-wrap
			     (z:invoke-open-unit-form-unit expr)))
		     (vars (map z:varref-var
			     (z:invoke-open-unit-form-variables expr))))
		(cond
		  ((null? name-spec)
		    `(#%invoke-open-unit ,unit))
		  ((not name-spec)
		    `(#%invoke-open-unit ,unit #f ,@vars))
		  ((symbol? name-spec)
		    `(#%invoke-open-unit ,unit ,name-spec ,@vars))
		  (else
		    (z:interface:internal-error name-spec
		      "given as name-spec for invoke-open-unit"))))]

	    [(z:interface-form? expr)
	      (let ((vars (z:interface-form-variables expr)))
		(for-each check-for-keyword vars)
		`(#%interface ,(map mv-wrap
				 (z:interface-form-super-exprs expr))
		   ,@(map read->raw vars)))]

	    [(z:class*/names-form? expr)
	      `(#%class*/names
		 (,(z:binding-var (z:class*/names-form-this expr))
		   ,(z:binding-var (z:class*/names-form-super-init expr)))
		 ,(mv-wrap (z:class*/names-form-super-expr expr))
		 ,(map mv-wrap (z:class*/names-form-interfaces expr))
		 ,(paroptarglist->ilist (z:class*/names-form-init-vars expr))
		 ,@(map
		     (lambda (clause)
		       (cond
			 ((z:public-clause? clause)
			   `(public
			      ,@(map (lambda (internal export expr)
				       `((,(z:binding-var internal)
					   ,(read->raw export))
					  ,(mv-wrap expr)))
				  (z:public-clause-internals clause)
				  (z:public-clause-exports clause)
				  (z:public-clause-exprs clause))))
			 ((z:private-clause? clause)
			   `(private
			      ,@(map (lambda (internal expr)
				       `(,(z:binding-var internal)
					  ,(mv-wrap expr)))
				  (z:private-clause-internals clause)
				  (z:private-clause-exprs clause))))
			 ((z:inherit-clause? clause)
			   `(inherit
			      ,@(map (lambda (internal inherited)
				       `(,(z:binding-var internal)
					  ,(read->raw inherited)))
				  (z:inherit-clause-internals clause)
				  (z:inherit-clause-imports clause))))
			 ((z:rename-clause? clause)
			   `(rename
			      ,@(map (lambda (internal import)
				       `(,(z:binding-var internal)
					  ,(read->raw import)))
				  (z:rename-clause-internals clause)
				  (z:rename-clause-imports clause))))
			 ((z:sequence-clause? clause)
			   `(sequence
			      ,@(map annotate
				  (z:sequence-clause-exprs clause))))))
		     (z:class*/names-form-inst-clauses expr)))]

	    [else
	      (print-struct #t)
	      (z:interface:internal-error
		expr
		(format "aries:annotate: unknown object to annotate, ~a~n" expr))])))

      (define transform
	(lambda (port offset file)
	  (let ([reader (z:read port
			  (z:make-location 1 1 offset file))])
	    (let read-loop ([exprs null])
	      (let ([expr (reader)])
		'(printf "expr: ~s~n" expr)
		(if (z:eof? expr)
		  (apply values (reverse exprs))
		  (let* ([expanded (z:scheme-expand expr)]
			  [_ '(printf "expanded: ~s~n" expanded)]
			  [annotated (annotate expanded)])
		    '(begin ((global-defined-value 'pretty-print) annotated)
		       (newline))
		    (read-loop (cons annotated exprs))))))))))
