;;
;; $Id: connect.ss,v 1.19 1998/01/27 21:54:09 robby Exp $
;;

  (unit/sig mred:connections^

    (import [wx : wx^]
	    [mred:constants : mred:constants^]
	    [mzlib:function : mzlib:function^]
            [mred : mred:testable-window^])

    (mred:debug:printf 'invoke "mred:connections@")

    (define make-connections-frame%
      (lambda (%)
	(class-asi %
	  (rename [super-show show])
	  (public
	    [shown #f]
	    [show (lambda (x) 
		    (set! shown x)
		    (super-show x))]
	    [on-frame-active void]
	    [active-edit
	     (lambda ()
	       (let ([c (active-canvas)])
		 (and c
		      (send c get-media))))]
	    [active-canvas
	     (let ([current-active #f])
	       (case-lambda
		[() current-active]
		[(new-active) (set! current-active new-active)]))]))))

    (define find-my-frame
      (lambda (parent)
	(let loop ([obj parent])
	  (cond
	    [(or (is-a? obj wx:frame%)
		 (is-a? obj wx:dialog-box%))
	      obj]
	    [(null? obj) #f]
	    [else (loop (send obj get-parent))]))))

    (define make-connections-media-canvas%
      (lambda (%)
	(class % (parent [x -1] [y -1]
			 [width -1]
			 [height -1]
			 [name ""]
			 [style 0]
			 [scrolls-per-page 100]
			 [init-buffer null])
	  (rename [super-set-media set-media]
		  [super-on-set-focus on-set-focus]
		  [super-on-kill-focus on-kill-focus])
	  (public
	    [frame (find-my-frame parent)])

	  (rename [super-show show])
	  (public
	    [shown #f]
	    [show (lambda (x)
		    (set! shown x)
		    (super-show x))])

	  (inherit get-media is-focus-on?)
  
	  (private
	    [update-active-canvas
	     (lambda (m)
	       (when (and (not (null? m))
			  (or (not (ivar m active-canvas))
			      (is-focus-on?)))
		 (send m set-active-canvas this)))])
	  (public
	    [edit-renamed
	     (lambda (name)
	       (void))]
	    [edit-modified
	     (lambda (modified?)
	       (void))]
	    [set-media
	     (opt-lambda (new [redraw? #t])
	       (let ([old (get-media)])
		 (update-active-canvas new)
		 (unless (null? old)
		   (send old on-kill-focus)
		   (send old remove-canvas this)))
	       (super-set-media new redraw?)
	       (unless (null? new)
		 (send new add-canvas this)
		 (when (is-focus-on?)
		   (send new on-set-focus))))]
	    [on-kill-focus
	     (lambda ()
	       (let ([m (get-media)])
		 (unless (null? m)
		   (send m on-kill-focus)))
	       (super-on-kill-focus))]
	    [on-set-focus
	     (lambda ()
	       (mred:debug:printf 'matthew "connect-canvas::on-set-focus~n")
	       (when frame
		 (send frame active-canvas this)
		 (mred:debug:printf 'matthew "connect-canvas::on-set-focus.2~n")
		 (send frame on-frame-active)
		 (mred:debug:printf 'matthew "connect-canvas::on-set-focus.3~n"))
	       (let ([m (get-media)])
		 (unless (null? m)
		   (mred:debug:printf 'matthew "connect-canvas::on-set-focus.4~n")
		   (send m set-active-canvas this)
		   (send m on-set-focus)
		   (mred:debug:printf 'matthew "connect-canvas::on-set-focus.5~n")))
	       (super-on-set-focus))])
	  (sequence
	    (super-init parent x y width height name style scrolls-per-page init-buffer)
	    (update-active-canvas init-buffer)))))

    (define make-connections-panel%
      (lambda (%)
	(class % (parent . args)
	  (public
	    [frame (find-my-frame parent)])
	  (sequence
	    (apply super-init parent args)))))

    (define make-connections-media-buffer%
      (lambda (%)
	(class-asi %
	  (rename [super-set-modified set-modified]
		  [super-set-filename set-filename])
	  (public
	    [on-set-focus (lambda () (void))]
	    [on-kill-focus (lambda () (void))]
	    [canvases null]
	    [active-canvas #f]
            [set-filename
             (opt-lambda (name [temp? #f])
               (super-set-filename name temp?)
               (for-each (lambda (canvas) (send canvas edit-renamed name))
                         canvases))]
            [set-modified
             (lambda (modified?)
               (super-set-modified modified?)
               (for-each (lambda (canvas) (send canvas edit-modified modified?))
                         canvases))]
	    [get-canvas
	     (lambda ()
	       (or active-canvas
		   (and (not (null? canvases))
			(car canvases))))]
	    [get-frame
	     (lambda ()
	       (let ([c (get-canvas)])
		 (and c
		      (ivar c frame))))]
	    [set-active-canvas
	     (lambda (new-canvas)
	       (set! active-canvas new-canvas))]
	    [add-canvas
	     (lambda (new-canvas)
	       (unless (member new-canvas canvases)
		 (set! canvases (cons new-canvas canvases))))]
	    [remove-canvas
	     (lambda (old-canvas)
	       (when (eq? old-canvas active-canvas)
		 (set! active-canvas #f))
	       (set! canvases (mzlib:function:remq old-canvas canvases)))]))))

    (define connections-panel%  (make-connections-panel% mred:testable-panel%))
    (define connections-frame%  (make-connections-frame% mred:testable-frame%))
    (define connections-dialog-box%   (make-connections-frame% mred:testable-dialog-box%))
    (define connections-media-canvas% (make-connections-media-canvas% mred:testable-media-canvas%))

    (define connections-media-edit%       (make-connections-media-buffer% wx:media-edit%)) 
    (define connections-media-pasteboard% (make-connections-media-buffer% wx:media-pasteboard%))

  )
