;;;
(require-library "sparams.ss" "backward")
(require-relative-library "userspcs.ss")
(require-library "ariess.ss" "cogen")
(require-library "cores.ss")
(require-library "pconvers.ss")
(require-library "zsigs.ss" "zodiac")
(require-library "sigs.ss" "zodiac")

;; extend structs with a parsing constructor
(define-macro define-struct/parse
  (lambda (str fields)
    (unless (symbol? str)
      (error 'define-struct/parse "no super structs allowed"))
    (let* ([first car]
	   [second cadr]
	   [second-name 'cadr]
	   [third caddr]
	   [defn (expand-defmacro `(#%define-struct ,str ,fields))]
	   [_ (unless (and (pair? defn)
			   (eq? (car defn) '#%define-values))
		(error 'define-struct/parse "expand-defmacro didn't return expected value: ~s~n" defn))]
	   [bindings (second defn)]
	   [exp (third defn)]
	   [make-parse (string->symbol (string-append "make-" (symbol->string str) "/parse"))]
	   [maker-name (second bindings)]
	   [parser
	    `(lambda (inits)
	       (apply ,maker-name
		      (map (lambda (field)
			     (let ([m (assq field inits)])
			       (unless m
				 (error ',make-parse "no binding for: ~a" field))
			       (unless (= (length m) 2)
				 (error ',make-parse "malformed binding: ~a" m))
			       (,second-name m)))
			   ',fields)))])
      `(define-values ,(cons make-parse bindings)
	 (call-with-values (lambda () ,exp)
			   (lambda bindings (apply values (cons ,parser bindings))))))))

(define-signature userspace:basis^
  (build-parameterization
   settings
   get-default-setting

   error-display/debug-handler
   current-vocabulary
   current-setting
   intermediate-values-during-load
   bottom-escape-handler

   drscheme-print

   INITIAL-LINE
   INITIAL-COLUMN
   INITIAL-OFFSET
   
   format-source-loc
   
   primitive-eval
   primitive-load
   syntax-checking-primitive-eval

   process/zodiac
   process/no-zodiac

   process-file/zodiac
   process-file/no-zodiac
   process-sexp/zodiac
   process-sexp/no-zodiac

   (struct process-finish (error?))

   level->number
   level-symbols
   level-strings

   copy-setting
   (struct setting (use-zodiac?
		    vocabulary-symbol
		    case-sensitive?
		    allow-set!-on-undefined?
		    unmatched-cond/case-is-error?
		    allow-improper-lists?
		    sharing-printing?
		    abbreviate-cons-as-list?
		    signal-undefined
		    signal-not-boolean
		    eq?-only-compares-symbols?
		    <=-at-least-two-args
		    disallow-untagged-inexact-numbers
		    print-tagged-inexact-numbers
		    whole/fractional-exact-numbers
		    printing))
   make-setting/parse
   r4rs-style-printing?
   find-setting-name))

(define-signature drscheme:interface^ 
  ((open zodiac:interface^)
   set-zodiac-phase
   (struct zodiac-exn (message start-location end-location type))))

(define-signature drscheme:exported-zodiac^ zodiac:system^)

(define-signature drscheme:zodiac^
  ((open drscheme:exported-zodiac^)
   current-vocabulary-symbol))

(define-signature userspace:basis-import^
  (in-mzscheme?))
