"The library provides access to the system's file directories. 

Assume and export: 
  (struct dir (name dirs files))
  (struct fil (name size contents))

A directory is 
  (make-dir symbol (listof directory) (listof file))
A file is 
  (make-fil symbol)

Export: 
  create-dir : path-name -> directory."

(require "error-lib.ss")

(define-signature dirS
  ((struct dir (name dirs files))
   (struct fil (name size contents))
   create-dir))
  
(define dirU
  (unit/sig dirS
    (import errorS plt:userspace^)
  
    (define-struct dir (name dirs files))
  
    (define-struct fil (name size contents))
  
    ;; create-dir : path -> directory
    (define (create-dir a-path)
      (check-arg 'create-dir (string? a-path) "string" "first" a-path)
      (if (directory-exists? a-path)
	  (car (explore (list a-path)))
	  (error 'create-dir "not a directory: ~e" a-path)))
  
    ;; explore : (listof directory-names) -> (listof directory)
    (define (explore dirs)
      (map (lambda (d) 
	     (let-values ([(fs ds) (pushd d directory-files&directories)]) 
	       (make-dir (string->symbol (my-split-path d))
		 (explore (map (lambda (x) (build-path d x)) ds))
		 (map make-fil 
		   (map string->symbol fs)
		   (map file-size
		     (map (lambda (x) (build-path d x)) fs))
		   (map (lambda (x) null) fs)))))
	dirs))
  
    (define (my-split-path d)
      (let-values ([(base name mbd?) (split-path d)])
	(if (string? base) name d)))
	  
	       
  
    ;; pushd : directory-name ( () -> X ) -> X
    (define (pushd d f)
      (let ((current (current-directory)))
	(dynamic-wind
	  (lambda () (current-directory d))
	  (lambda () (f))
	  (lambda () (current-directory current)))))
  
    ;; directory-files&directories : 
    ;;  () -> (values (listof file-names) (listof directory-names))
    (define (directory-files&directories)
      (let ((contents (directory-list)))
	(values (filter file-exists? contents)
	  (filter directory-exists? contents))))
  
    ;; filter: (X -> bool) (listof X) -> (listof X)
    ;;  (define (filter p? l)
    ;;    (foldr (lambda (fst rst) (if (p? fst) (cons fst rst) rst)) '() l))
    ))

(compound-unit/sig
  (import (PLT : plt:userspace^))
  (link
    (XXX : dirS (dirU ERR PLT))
    (ERR  : errorS (errorU)))
  (export (open XXX)))
