"The library provides operations for dealing with primitive HTML documents. 

An _annotation_ is a string with 
     #\< as the first and
     #\> as the last char.

An _end annotation is a string with 
     #\< as the first, 
     #\/ as the second, and 
     #\> as the last char.

It provides: 
  annotation? 
  end-annotation
  write-file."

(define-signature docS (annotation? end-annotation write-file))

(define docU 
  (unit/sig docS (import errorS plt:userspace^)

    (printf "hello world~n")

    ;; mk-annotation : 
    ;;  (str number[str-lenght] -> bool) -> (any-value -> bool)
    (define (mk-annotation test)
      (lambda (val)
	(and (symbol? val) 
	  (let ((str (symbol->string val)))
	    (test str (sub1 (string-length str)))))))

    ;; annotation? : any-value -> bool
    (define annotation? 
      (mk-annotation (lambda (str E);; str : string, E : integer
		       (and (> E 1) 
			 (eqv? (string-ref str 0) #\<)
			 (eqv? (string-ref str E) #\>)))))

    ;; end-annotation? : any-value -> bool
    (define end-annotation? 
      (mk-annotation (lambda (str E);; str : string, E : integer 
		       (and (> E 2)
			 (eqv? (string-ref str 0) #\<)
			 (eqv? (string-ref str 1) #\/)
			 (eqv? (string-ref str E) #\>)))))

    ;; end-annotation : annotation -> end-annotation
    (define (end-annotation ann)
      (check-arg 'end-annotation (annotation? ann) "annotation (str)" "first" ann)
      (let ((str (symbol->string ann))) 
	(string->symbol
	  (string-append "</" (substring str 1 (string-length str))))))

    ;; line-breaking? : any-value -> bool
    (define (line-breaking? x)
      (and (annotation? x) (memq x LNBRK) #t))

    (define LNBRK 
      (let ((x (list '<html> '<title> '<body>)))
	(append x (map end-annotation x))))

    (printf "ln break is defined: ~a~n" LNBRK)

    ;; write-file : list-of-(symbols|numbers) -> void
    (define (write-file  los)
      (check-arg 'write-file 
	(and (list? los) (andmap (lambda (x) (or (symbol? x) (number? x))) los))
	"list of symbols" "first" los)
      (let loop ((i 0) (los los))
	(if (null? los) (newline)
	  (let* ((wrd (first los))
		 (str (if (number? wrd) (number->string wrd) (symbol->string wrd)))
		 (j   (+ i (string-length str) 1)))
	    (cond
	      ((> i 80)              (newline) 
	       (loop 0 los))
	      ((line-breaking? wrd)  (newline) (display wrd) (newline)
	       (loop j (rest los)))
;	    ((end-annotation? wrd) (display wrd) (newline)
;	     (loop 0 (rest los)))
;	    ((annotation? wrd)     (unless (= i 0) (newline))
;	                           (display wrd) (display #\SPACE) 
;	     (loop j (rest los)))
	      (else                  (display wrd) (display #\SPACE)
		(loop j (rest los))))))))

    ))

(compound-unit/sig
  (import (PLT : plt:userspace^))
  (link
    (ERR  : errorS (errorU))
    (DRAW : docS (docU ERR PLT)))
  (export (open DRAW)))
