;; simple functions to get figures drawn

(require "draw-sig.ss")
(require "error-lib.ss")

(define drawU
  (unit/sig drawS
    (import errorS plt:userspace^)
  
    (define (make-error tag)
      (lambda x
	(error "evaluate (start <num> <num>) first")))  

    (define %old-sr (make-error '%oldsr))
  
    (define %old-li (make-error '%old-li))
  
    (define draw-solid-disk (make-error 'draw-solid-disk))
    (define clear-solid-disk (make-error 'clear-solid-disk))

    (define draw-solid-rect (make-error 'draw-solid-rect))
    (define clear-solid-rect (make-error 'clear-solid-rect))
  
    (define draw-solid-line (make-error 'draw-solid-line))
    (define clear-solid-line (make-error 'clear-solid-line))

    (define clear-all (make-error 'clear-all))

    (define (make-true f) (lambda x (apply f x) #t))

    (define sleep-for-a-while (make-true sleep))

    ;; i wish i could abstract these functions ...
    (define (make-line name f)
      (make-true
	(lambda x
	  (check-arity name 2 x)
	  (apply (lambda (p1 p2 . c)
		   (check-arg name (posn? p1) "posn" "first" p1)
		   (check-arg name (posn? p2) "posn" "second" p2)
		   (f p1 p2 (if (null? c)
				BLACK
				(let ((c (car c)))
				  (check-arg name (rgb? c) "color" "third" c)
				  c))))
	    x))))

    (define (make-rect name f)
      (make-true
	(lambda x
	  (check-arity name 3 x)
	  (apply (lambda (p w h . c)
		   (check-arg name (posn? p) "posn" "first" p)
		   (check-arg name (and (integer? w) (> w 0)) "positive integer" "second" w)
		   (check-arg name (and (integer? h) (> h 0)) "positive integer" "third" h)
		   (f p w h (if (null? c)
				BLACK
				(let ((c (car c)))
				  (check-arg name (rgb? c) "color" "fourth" c)
				  c))))
	    x))))
	
    (define (make-circle name f)
      (make-true
	(lambda x
	  (check-arity name 2 x)
	  (apply (lambda (p r . c)
		   (check-arg name (posn? p) "posn" "first" p)
		   (check-arg name (and (integer? r) (> r 0)) "positive integer" "second" r)
		   (let ((d (* r 2))
			 (c (if (null? c) BLACK (car c))))
		     (check-arg name (rgb? c) "color" "third" c)
		     (f (make-posn (- (posn-x p) r) (- (posn-y p) r)) d d c)))
	    x))))

    (define (start WIDTH HEIGHT)
      (check-arg 'start (and (integer? WIDTH) (> WIDTH 0)) "positive integer" "first" WIDTH)
      (check-arg 'start (and (integer? HEIGHT) (> HEIGHT 0)) "positive integer" "second" HEIGHT)
      ;; --- 
      (open-graphics)
      (set! %old-sr draw-solid-rectangle)
      (set! %old-li draw-line)
      (let ((current-window (open-viewport "Canvas" WIDTH HEIGHT)))
	(set! @VP current-window)
	(set! clear-all (clear-viewport current-window))
	;; --- 
	(set! draw-solid-line (make-line 'draw-solid-line (draw-line current-window)))
	(set! clear-solid-line (make-line 'clear-solid-line (clear-line current-window)))
	;; --- 
	(set! draw-solid-rect (make-rect 'draw-solid-rect (draw-solid-rectangle current-window)))
	(set! clear-solid-rect (make-rect 'clear-solid-rect (clear-solid-rectangle current-window)))
	;; --- 
	(set! draw-solid-disk (make-circle 'draw-solid-disk (draw-solid-ellipse current-window)))
	(set! clear-solid-disk (make-circle 'clear-solid-disk (clear-solid-ellipse current-window)))))
  
    (define (stop)
      (set! draw-solid-rect %old-sr)
      (set! draw-solid-line %old-li)
      (close-graphics))
  
    (define @VP #f)

    (define WHITE (make-rgb 1 1 1))
    (define YELLOW (make-rgb 1 1 0))
    (define RED (make-rgb 1.0 0 0))
    (define GREEN (make-rgb 0 1.0 0))
    (define BLUE (make-rgb 0 0 1.0))
    (define BLACK (make-rgb 0 0 0))
  
    ))

(compound-unit/sig
  (import (PLT : plt:userspace^))
  (link
    (DRAW : drawS (drawU ERR PLT))
    (ERR  : errorS (errorU)))
  (export (open DRAW)))
