(require "error-lib.ss")
(require "draw-lib.ss")

(define-signature hangmanS (hangman-repl))

(define hangmanU
  (unit/sig hangmanS (import errorS drawS (plt : plt:userspace^))

    (define LETTERS '(a b c d e f g h i j k l m n o p q r s t u v w x y z _))

    (define LET-PER-WORD 3)

    (define PARTS (vector 'right-leg 'left-leg 'left-arm 'right-arm 'body 'head 'noose))
    (define NUMBER-OF-GUESSES (vector-length PARTS))
    (define pieces-left (sub1 NUMBER-OF-GUESSES))

    ;; hangman-repl : word word reveal draw-next-part -> void
    ;; effect: query user for a guess, apply reveal until hung or (equal? wrd status)
    (define (hangman-repl wrd status reveal draw-next-part)
      ;; -- check args
      ; the following checks assume we know how the student represented the word; NO!
      ; (check-arg proc (and (list? wrd) (andmap symbol? wrd)) "word (list-of-a..z)" 'first wrd)
      ; (check-arg proc (and (list? status) (andmap symbol? status)) "word (list-of-a..z)" 'first status)
      (check-proc 'hangman-repl reveal         3 '3rd "3 arguments")
      (check-proc 'hangman-repl draw-next-part 1 '4th "1 argument")      
      ;; ---
      (stop) (start 200 400)
      (set! pieces-left (sub1 NUMBER-OF-GUESSES))
      (draw-next-part (vector-ref PARTS pieces-left))
      (let repl ((status status) (guesses null))
	(let ((guess (begin (printf "Your next letter please: ") (read))))
	  (cond 
	    ((eq? guess 'quit) (void))
	    ((not (member guess LETTERS))
	     (printf "~a is not a letter. Try again!~n" guess)
	     (repl status guesses))
	    ((member guess guesses)
	     (printf "You're losing your memory.")
	     (repl status guesses))
	    (else (let ((new (reveal wrd status guess)))
		    (cond
		      ((equal? new wrd) (printf "You won. Good bye."))
		      ((equal? status new) 
		       (set! pieces-left (- pieces-left 1))
		       (draw-next-part (vector-ref PARTS pieces-left))
		       (if (= pieces-left 0)
			   (printf "This is the end my only friend.~n")
			   (repl new (cons guess guesses))))
		      (else
			(printf "Good. Here is what we know: ~a~n" new)
			(repl new (cons guess guesses))))))))))))

(compound-unit/sig (import (PLT : plt:userspace^))
  (link
    [ERR   : errorS (errorU)]
    [DRAW  : drawS  ((require-unit/sig "draw-lib.ss") PLT)]
    [HANG  : hangmanS (hangmanU ERR DRAW PLT)])
  (export (open DRAW) (open HANG)))
